/*
 * Copyright (c) 2000, 2001 University of Utah and the Flux Group.
 * All rights reserved.
 * 
 * This file is part of Clack.
 * 
 * Clack is free software, also known as "open source."  You can redistribute
 * it and/or modify it under the terms of the GNU General Public License
 * (GPL), version 2, as published by the Free Software Foundation (FSF).
 * 
 * Clack is distributed in the hope that it will be useful, but WITHOUT ANY
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 * FOR A PARTICULAR PURPOSE.  See the GPL for more details.  You should have
 * received a copy of the GPL along with Clack; see the file COPYING.  If
 * not, write to the FSF, 59 Temple Place #330, Boston, MA 02111-1307, USA.
 * 
 * Clack is modelled on and includes some code fragments from MIT's Click
 * system (http://www.pdos.lcs.mit.edu/click/) which is copyright (c)
 * 1999-2000 Massachusetts Institute of Technology and also GPL'ed.
 * 
 * Users are requested, but not required, to send to csl-dist@cs.utah.edu any
 * improvements that they make and grant Univ. of Utah redistribution rights.
 */

#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include <errno.h>
#include <clack/c/packet.h>
#include <clack/c/packet_eth.h>
#include <clack/c/if_arp.h>
#include <arpa/inet.h>
#include <clack/c/arp_responder.h>

/*
 * Imported:  push_in1 - Output link for Ether/Arp reply packets.
 *            push_in2 - Output link for failed arp lookups.
 */
extern clack_error_t	push_in1(clack_packet_t *p);
extern clack_error_t	push_in2(clack_packet_t *p);

/*
 * Also imported, the arp responder table is hardwired with the set of
 * IP/Ether pairs we can respond for. The incoming table is in text
 * format, which we have to convert to internal representation. The
 * table *MUST* have a null entry to mark the end of the table.
 */
extern struct arp_descriptor_entry	arp_descriptor_table[];

/*
 * Exported:  push - Incoming Ether/Arp request packets.
 */
clack_error_t		push(clack_packet_t *p);

/*
 * Locally allocated table. 
 */
struct arp_mapping_entry {
	struct in_addr	am_ip;
	unsigned char	am_eth[ETHER_ADDR_SIZE];
};
static struct arp_mapping_entry	       *arp_mapping_table;
static int				arp_mapping_table_count;

/*
 * Initialize
 */
clack_error_t
init(void) 
{
	int				count = 0;
	struct arp_descriptor_entry	*entry;
	struct arp_mapping_entry	*mapping;

	entry = arp_descriptor_table;
	while (entry->ip) {
		count++;
		entry++;
	}
	arp_mapping_table_count = count;
	printf("arp_responder(init): %d entries\n", count);

	if ((arp_mapping_table = (struct arp_mapping_entry *)
	     malloc(count * sizeof(struct arp_mapping_entry))) == NULL) {
		printf("arp_responder(init): Count not allocate table\n");
		return ENOMEM;
	}

	/*
	 * Copy text entries into real arp mapping table.
	 */
	entry   = arp_descriptor_table;
	mapping = arp_mapping_table;
	while (entry->ip) {
		int rc = inet_aton(entry->ip, &mapping->am_ip);
		if (!rc)
			return 1;  // inet_aton returns 1 for success
		memcpy(mapping->am_eth,
		       entry->ethaddr, sizeof(entry->ethaddr));
		
		entry++;
		mapping++;
	}

	return 0;
}

clack_error_t
push(clack_packet_t *p)
{
	struct ether_header *e  = (struct ether_header *) clack_packet_data(p);
	struct ether_arp    *ea = (struct ether_arp *) (e + 1);
	struct ether_arp    *r;
	clack_packet_t	    *reply;
	struct arp_mapping_entry *mapping;
	int		    i;

	/*
	 * Look for the entry in the table. Push it out the failure side
	 * if its not something we can arp respond for.
	 */
	mapping = arp_mapping_table;
	for (i = 0; i < arp_mapping_table_count; i++, mapping++) {
		if (mapping->am_ip.s_addr ==
		    *((unsigned int *) &(ea->arp_tpa)))
			break;
	}
	if (i == arp_mapping_table_count) {
		push_in2(p);
		return 0;
	}

	/*
	 * Found a mapping.
	 */
	// ToDo: cleanup hack to pad packet out to 60 bytes long
	reply = clack_packet_alloc_padded(18+sizeof (struct ether_arp),
					  sizeof (struct ether_header));
	  
	// assert(reply); // ToDo: kill all asserts

	r = (struct ether_arp *) clack_packet_data(reply);
	r->ea_hdr.ar_hrd = htons(ARPHRD_ETHER);
	r->ea_hdr.ar_pro = htons(ETHERTYPE_IP);
	r->ea_hdr.ar_hln = ETHER_ADDR_SIZE;
	r->ea_hdr.ar_pln = sizeof (struct in_addr);
	r->ea_hdr.ar_op  = htons(ARPOP_REPLY);
	memcpy(r->arp_tha, ea->arp_sha, ETHER_ADDR_SIZE);
	memcpy(r->arp_tpa, ea->arp_spa, sizeof (struct in_addr));
	memcpy(r->arp_sha, mapping->am_eth, ETHER_ADDR_SIZE);
	memcpy(r->arp_spa, ea->arp_tpa, sizeof (struct in_addr));
          
	// ToDo: should arp reply go to address in ether
	// header or arp request?
	add_eth_hdr(reply,ea->arp_sha,mapping->am_eth,ETHERTYPE_ARP);
	push_in1(reply);

	clack_packet_release(p);          
	return 0;
}
